@@ -109,6 +109,9 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
     else if (tc->instance->debugserver)
         MVM_breakpoint_instrument(tc, static_frame);
     else
+        /* XXX uninstrumenting is currently turned off, due to multithreading
+         * woes. If you add an instrumentation that has to be "turned off"
+         * again at some point, a solution for this problem must be found. */
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
@@ -398,7 +401,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
          * its reference count incremented; just ensure that it is based on the
          * correct static frame (compare on bytecode address to cope with
          * nqp::freshcoderef). */
-        if (outer->static_info->body.orig_bytecode != static_frame->body.outer->body.orig_bytecode) {
+        if (MVM_UNLIKELY(outer->static_info->body.orig_bytecode != static_frame->body.outer->body.orig_bytecode)) {
             char *frame_cuuid = MVM_string_utf8_encode_C_string(tc, static_frame->body.cuuid);
             char *frame_name;
             char *outer_cuuid = MVM_string_utf8_encode_C_string(tc, outer->static_info->body.cuuid);
@@ -551,6 +554,8 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     /* Initialize argument processing. */
     MVM_args_proc_init(tc, &frame->params, callsite, args);
 
+    MVM_jit_code_trampoline(tc);
+
     /* Update interpreter and thread context, so next execution will use this
      * frame. */
     tc->cur_frame = frame;
@@ -579,7 +584,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                     redo_state:
                     switch (state_act) {
                     case 0:
-                        if (!frame->code_ref)
+                        if (MVM_UNLIKELY(!frame->code_ref))
                             MVM_exception_throw_adhoc(tc,
                                 "Frame must have code-ref to have state variables");
                         state = ((MVMCode *)frame->code_ref)->body.state_vars;
@@ -883,6 +888,16 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
 
     /* Switch back to the caller frame if there is one. */
     if (caller && returner != tc->thread_entry_frame) {
+
+       if (tc->jit_return_address != NULL) {
+            /* on a JIT frame, exit to interpreter afterwards */
+            MVMJitCode *jitcode = returner->spesh_cand->jitcode;
+            MVM_jit_code_set_current_position(tc, jitcode, returner, jitcode->exit_label);
+            /* given that we might throw in the special-return, act as if we've
+             * left the current frame (which is true) */
+            tc->jit_return_address = NULL;
+        }
+
         tc->cur_frame = caller;
         tc->current_frame_nr = caller->sequence_nr;
 
@@ -992,6 +1007,7 @@ typedef struct {
     MVMFrame  *frame;
     MVMuint8  *abs_addr;
     MVMuint32  rel_addr;
+    void      *jit_return_label;
 } MVMUnwindData;
 static void mark_unwind_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *worklist) {
     MVMUnwindData *ud  = (MVMUnwindData *)frame->extra->special_return_data;
@@ -1002,11 +1018,12 @@ static void continue_unwind(MVMThreadContext *tc, void *sr_data) {
     MVMFrame *frame    = ud->frame;
     MVMuint8 *abs_addr = ud->abs_addr;
     MVMuint32 rel_addr = ud->rel_addr;
+    void *jit_return_label = ud->jit_return_label;
     MVM_free(sr_data);
-    MVM_frame_unwind_to(tc, frame, abs_addr, rel_addr, NULL);
+    MVM_frame_unwind_to(tc, frame, abs_addr, rel_addr, NULL, jit_return_label);
 }
 void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_addr,
-                         MVMuint32 rel_addr, MVMObject *return_value) {
+                         MVMuint32 rel_addr, MVMObject *return_value, void *jit_return_label) {
     while (tc->cur_frame != frame) {
         MVMFrame *cur_frame = tc->cur_frame;
         if (cur_frame->static_info->body.has_exit_handler &&
@@ -1042,6 +1059,7 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
                 ud->frame = frame;
                 ud->abs_addr = abs_addr;
                 ud->rel_addr = rel_addr;
+                ud->jit_return_label = jit_return_label;
                 if (return_value)
                     MVM_exception_throw_adhoc(tc, "return_value + exit_handler case NYI");
                 MVM_frame_special_return(tc, cur_frame, continue_unwind, NULL, ud,
@@ -1065,13 +1083,18 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
         *tc->interp_cur_op = abs_addr;
     else if (rel_addr)
         *tc->interp_cur_op = *tc->interp_bytecode_start + rel_addr;
+
+    if (jit_return_label) {
+        MVM_jit_code_set_current_position(tc, tc->cur_frame->spesh_cand->jitcode, tc->cur_frame, jit_return_label);
+    }
+
     if (return_value)
         MVM_args_set_result_obj(tc, return_value, 1);
 }
 
 /* Gets a code object for a frame, lazily deserializing it if needed. */
 MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
-    if (REPR(code)->ID != MVM_REPR_ID_MVMCode)
+    if (MVM_UNLIKELY(REPR(code)->ID != MVM_REPR_ID_MVMCode))
         MVM_exception_throw_adhoc(tc, "getcodeobj needs a code ref");
 
     if (!code->body.code_object) {
@@ -1080,7 +1103,7 @@ MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
             MVMObject *resolved;
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, sf->body.cu,
                 sf->body.code_obj_sc_dep_idx - 1);
-            if (sc == NULL)
+            if (MVM_UNLIKELY(sc == NULL))
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
 
@@ -1099,7 +1122,7 @@ MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
 void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
     MVMCode *code_obj = (MVMCode *)code;
     MVMFrame *captured;
-    if (REPR(code)->ID != MVM_REPR_ID_MVMCode)
+    if (MVM_UNLIKELY(REPR(code)->ID != MVM_REPR_ID_MVMCode))
         MVM_exception_throw_adhoc(tc,
             "Can only perform capturelex on object with representation MVMCode");
     MVMROOT(tc, code, {
@@ -1144,7 +1167,7 @@ MVMObject * MVM_frame_takeclosure(MVMThreadContext *tc, MVMObject *code) {
     MVMCode *closure;
     MVMFrame *captured;
 
-    if (REPR(code)->ID != MVM_REPR_ID_MVMCode)
+    if (MVM_UNLIKELY(REPR(code)->ID != MVM_REPR_ID_MVMCode))
         MVM_exception_throw_adhoc(tc,
             "Can only perform takeclosure on object with representation MVMCode");
 
@@ -1246,7 +1269,7 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
             MVMLexicalRegistry *entry;
             MVM_HASH_GET(tc, lexical_names, name, entry)
             if (entry) {
-                if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
+                if (MVM_LIKELY(cur_frame->static_info->body.lexical_types[entry->value] == type)) {
                     MVMRegister *result = &cur_frame->env[entry->value];
                     if (type == MVM_reg_obj && !result->o)
                         MVM_frame_vivify_lexical(tc, cur_frame, entry->value);
@@ -1263,7 +1286,7 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
         }
         cur_frame = cur_frame->outer;
     }
-    if (type != MVM_reg_obj) {
+    if (MVM_UNLIKELY(type != MVM_reg_obj)) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
@@ -1314,7 +1337,7 @@ MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *
 /* Finds a lexical in the outer frame, throwing if it's not there. */
 MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name) {
     MVMRegister *r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
-    if (r)
+    if (MVM_LIKELY(r != NULL))
         return r->o;
     else {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -1432,7 +1455,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
     MVMuint64 last_time;
 
     MVMFrame *initial_frame = cur_frame;
-    if (!name)
+    if (MVM_UNLIKELY(!name))
         MVM_exception_throw_adhoc(tc, "Contextual name cannot be null");
     if (dlog) {
         c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -1450,39 +1473,37 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
          * knows where to find them */
         if (cand && cand->num_inlines) {
             if (cand->jitcode) {
-                void      **labels = cand->jitcode->labels;
-                void *return_label = cur_frame->jit_entry_label;
-                MVMJitInline *inls = cand->jitcode->inlines;
+                MVMJitCode *jitcode = cand->jitcode;
+                void * current_position = MVM_jit_code_get_current_position(tc, jitcode, cur_frame);
                 MVMint32 i;
-                if (return_label == NULL)
-                    MVM_oops(tc, "Return label is NULL!\n");
-                for (i = 0; i < cand->jitcode->num_inlines; i++) {
+
+                for (i = MVM_jit_code_get_active_inlines(tc, jitcode, current_position, 0);
+                     i < jitcode->num_inlines;
+                     i = MVM_jit_code_get_active_inlines(tc, jitcode, current_position, i+1)) {
+                    MVMStaticFrame *isf = cand->inlines[i].sf;
                     icost++;
-                    if (return_label >= labels[inls[i].start_label] && return_label <= labels[inls[i].end_label]) {
-                        MVMStaticFrame *isf = cand->inlines[i].sf;
-                        if ((lexical_names = isf->body.lexical_names)) {
-                            MVMLexicalRegistry *entry;
-                            MVM_HASH_GET(tc, lexical_names, name, entry);
-                            if (entry) {
-                                MVMuint16    lexidx = cand->inlines[i].lexicals_start + entry->value;
-                                MVMRegister *result = &cur_frame->env[lexidx];
-                                *type = cand->lexical_types[lexidx];
-                                if (vivify && *type == MVM_reg_obj && !result->o) {
-                                    MVMROOT3(tc, cur_frame, initial_frame, name, {
-                                        MVM_frame_vivify_lexical(tc, cur_frame, lexidx);
-                                    });
-                                }
-                                if (fcost+icost > 1)
-                                  try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
-                                if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
-                                    fflush(dlog);
-                                    MVM_free(c_name);
-                                    tc->instance->dynvar_log_lasttime = uv_hrtime();
-                                }
-                                *found_frame = cur_frame;
-                                return result;
+                    if ((lexical_names = isf->body.lexical_names)) {
+                        MVMLexicalRegistry *entry;
+                        MVM_HASH_GET(tc, lexical_names, name, entry);
+                        if (entry) {
+                            MVMuint16    lexidx = cand->inlines[i].lexicals_start + entry->value;
+                            MVMRegister *result = &cur_frame->env[lexidx];
+                            *type = cand->lexical_types[lexidx];
+                            if (vivify && *type == MVM_reg_obj && !result->o) {
+                                MVMROOT3(tc, cur_frame, initial_frame, name, {
+                                    MVM_frame_vivify_lexical(tc, cur_frame, lexidx);
+                                });
                             }
+                            if (fcost+icost > 1)
+                                try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
+                            if (dlog) {
+                                fprintf(dlog, "I %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+                                fflush(dlog);
+                                MVM_free(c_name);
+                                tc->instance->dynvar_log_lasttime = uv_hrtime();
+                            }
+                            *found_frame = cur_frame;
+                            return result;
                         }
                     }
                 }
@@ -1589,7 +1610,7 @@ MVMObject * MVM_frame_getdynlex(MVMThreadContext *tc, MVMString *name, MVMFrame
     MVMRegister *lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 1, &found_frame);
     MVMObject *result = NULL, *result_type = NULL;
     if (lex_reg) {
-        switch (type) {
+        switch (MVM_EXPECT(type, MVM_reg_obj)) {
             case MVM_reg_int64:
                 result_type = (*tc->interp_cu)->body.hll_config->int_box_type;
                 if (!result_type)
@@ -1671,13 +1692,12 @@ void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *valu
  * try to vivify anything - gets exactly what is there. */
 MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *name) {
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
-    if (lexical_names) {
+    if (MVM_LIKELY(lexical_names != NULL)) {
         MVMLexicalRegistry *entry;
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry)
             return &f->env[entry->value];
     }
-
     {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
@@ -1709,7 +1729,7 @@ MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMStrin
         MVMLexicalRegistry *entry;
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry) {
-            switch (f->static_info->body.lexical_types[entry->value]) {
+            switch (MVM_EXPECT(f->static_info->body.lexical_types[entry->value], MVM_reg_obj)) {
                 case MVM_reg_int64:
                     return MVM_STORAGE_SPEC_BP_INT;
                 case MVM_reg_num64:
@@ -1718,9 +1738,23 @@ MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMStrin
                     return MVM_STORAGE_SPEC_BP_STR;
                 case MVM_reg_obj:
                     return MVM_STORAGE_SPEC_BP_NONE;
+                case MVM_reg_int8:
+                    return MVM_STORAGE_SPEC_BP_INT8;
+                case MVM_reg_int16:
+                    return MVM_STORAGE_SPEC_BP_INT16;
+                case MVM_reg_int32:
+                    return MVM_STORAGE_SPEC_BP_INT32;
+                case MVM_reg_uint8:
+                    return MVM_STORAGE_SPEC_BP_UINT8;
+                case MVM_reg_uint16:
+                    return MVM_STORAGE_SPEC_BP_UINT16;
+                case MVM_reg_uint32:
+                    return MVM_STORAGE_SPEC_BP_UINT32;
+                case MVM_reg_uint64:
+                    return MVM_STORAGE_SPEC_BP_UINT64;
                 default:
                 {
-                    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+                    char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
                     char *waste[] = { c_name, NULL };
                     MVM_exception_throw_adhoc_free(tc, waste,
                         "Unhandled lexical type '%s' in lexprimspec for '%s'",
@@ -1883,7 +1917,7 @@ MVMObject * MVM_frame_resolve_invokee_spesh(MVMThreadContext *tc, MVMObject *inv
     }
     else {
         MVMInvocationSpec *is = STABLE(invokee)->invocation_spec;
-        if (is && is->code_ref_offset && IS_CONCRETE(invokee))
+        if (MVM_LIKELY(is && is->code_ref_offset && IS_CONCRETE(invokee)))
             return MVM_p6opaque_read_object(tc, invokee, is->code_ref_offset);
     }
     return tc->instance->VMNull;
