@@ -46,6 +46,8 @@ typedef struct {
     MVMuint32         cur_instr;
     MVMCallsite      *cur_call;
     MVMuint16         cur_arg;
+    MVMint32          acceptable_max_arity;
+    MVMint16          checkarity_seen;
     MVMCallsiteEntry  expected_named_arg;
     MVMuint16         remaining_args;
     MVMuint16         remaining_positionals;
@@ -54,7 +56,7 @@ typedef struct {
 } Validator;
 
 
-MVM_NO_RETURN static void fail(Validator *val, const char *msg, ...) MVM_FORMAT(printf, 2, 3) MVM_NO_RETURN_GCC;
+MVM_NO_RETURN static void fail(Validator *val, const char *msg, ...) MVM_FORMAT(printf, 2, 3) MVM_NO_RETURN_ATTRIBUTE;
 static void fail(Validator *val, const char *msg, ...) {
     va_list args;
 
@@ -392,12 +394,53 @@ static void validate_operands(Validator *val) {
 
             break;
         }
+        case MVM_OP_checkarity: {
+            validate_literal_operand(val, operands[0]);
+
+            validate_literal_operand(val, operands[1]);
+            val->acceptable_max_arity = GET_UI16(val->cur_op, -2);
+            val->checkarity_seen = 1;
+
+            break;
+        }
+        case MVM_OP_wval:
+        case MVM_OP_wval_wide: {
+            validate_reg_operand(val, operands[0]);
+            validate_literal_operand(val, operands[1]);
+            if (GET_UI16(val->cur_op, -2) >= val->cu->body.num_scs)
+                fail(val, MSG(val, "out of range SC index %u"), GET_UI16(val->cur_op, -2));
+            validate_literal_operand(val, operands[2]);
+            break;
+        }
 
         default: {
             int i;
 
-            for (i = 0; i < val->cur_info->num_operands; i++)
-                validate_operand(val, val->cur_info->operands[i]);
+            if (val->cur_mark[1] == 'p') {
+                /* First of all, bail out if no checkarity was seen yet. */
+                if (!val->checkarity_seen) {
+                    fail(val, MSG(val, "param op without checkarity op seen."));
+                }
+
+                /* For the p-marked ops, which is a subset of param_* ops,
+                 * we check the second argument against the value checkarity
+                 * checked against. */
+                for (i = 0; i < val->cur_info->num_operands; i++) {
+                    validate_operand(val, val->cur_info->operands[i]);
+
+                    /* This is the argument we want to check */
+                    if (i == 1) {
+                        MVMint16 value = GET_UI16(val->cur_op, -2);
+                        if (value > val->acceptable_max_arity) {
+                            fail(val, MSG(val, "tried to take arg number %d after checkarity with %d"), value, val->acceptable_max_arity);
+                        }
+                    }
+                }
+            }
+            else {
+                for (i = 0; i < val->cur_info->num_operands; i++)
+                    validate_operand(val, val->cur_info->operands[i]);
+            }
         }
     }
 }
@@ -620,13 +663,16 @@ void MVM_validate_static_frame(MVMThreadContext *tc,
     val->bc_size   = fb->bytecode_size;
     val->src_cur_op = fb->bytecode;
     val->src_bc_end = fb->bytecode + fb->bytecode_size;
-    val->labels    = MVM_calloc(fb->bytecode_size, 1);
+    val->labels    = MVM_calloc(1, fb->bytecode_size);
     val->cur_info  = NULL;
     val->cur_mark  = NULL;
     val->cur_instr = 0;
     val->cur_call  = NULL;
     val->cur_arg   = 0;
 
+    val->acceptable_max_arity  = 0;
+    val->checkarity_seen       = 0;
+
     val->expected_named_arg    = 0;
     val->remaining_positionals = 0;
     val->remaining_jumplabels  = 0;
@@ -670,26 +716,6 @@ void MVM_validate_static_frame(MVMThreadContext *tc,
     validate_branch_targets(val);
     validate_final_return(val);
 
-    /* Validation successful. Cache the located instruction offsets. */
-    fb->instr_offsets = val->labels;
-}
-
-
-/* Returns MVM_BC_ILLEGAL_OFFSET if the offset is out of range or does not
- * point to an op boundary. */
-MVMuint32 MVM_bytecode_offset_to_instr_idx(MVMThreadContext *tc,
-        MVMStaticFrame *static_frame, MVMuint32 offset) {
-    MVMuint8 *labels = static_frame->body.instr_offsets;
-    MVMuint32 i, idx = 0;
-
-    if (offset >= static_frame->body.bytecode_size
-            || (labels[offset] & MVM_BC_op_boundary) == 0)
-        return MVM_BC_ILLEGAL_OFFSET;
-
-    for (i = 0; i < offset; i++) {
-        if (labels[i] & MVM_BC_op_boundary)
-            idx++;
-    }
-
-    return idx;
+    /* Validation successful. Clear up instruction offsets. */
+    MVM_free(val->labels);
 }
