@@ -40,12 +40,16 @@ static MVMProfileCallNode *take_node(MVMThreadContext *tc, NodeWorklist *list) {
 }
 
 /* Adds an instruction to log an allocation. */
-static void add_allocation_logging(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+static void add_allocation_logging_at_location(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins, MVMSpeshIns *location) {
     MVMSpeshIns *alloc_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
     alloc_ins->info        = MVM_op_get_op(MVM_OP_prof_allocated);
     alloc_ins->operands    = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshOperand));
     alloc_ins->operands[0] = ins->operands[0];
-    MVM_spesh_manipulate_insert_ins(tc, bb, ins, alloc_ins);
+    MVM_spesh_manipulate_insert_ins(tc, bb, location, alloc_ins);
+}
+
+static void add_allocation_logging(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    add_allocation_logging_at_location(tc, g, bb, ins, ins);
 }
 
 static void add_nativecall_logging(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
@@ -109,9 +113,7 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             }
             case MVM_OP_invoke_o:
             case MVM_OP_param_rp_o:
-            case MVM_OP_param_op_o:
             case MVM_OP_param_rn_o:
-            case MVM_OP_param_on_o:
             case MVM_OP_param_sp:
             case MVM_OP_param_sn:
             case MVM_OP_newexception:
@@ -152,6 +154,19 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
                 add_allocation_logging(tc, g, bb, ins);
                 break;
             }
+            case MVM_OP_param_op_o:
+            case MVM_OP_param_on_o: {
+                /* These ops jump to a label if they "succeed", so the
+                 * allocation logging goes in another bb. */
+                MVMSpeshBB  *target   = ins->operands[2].ins_bb;
+                MVMSpeshIns *location = target->first_ins;
+                while (location && location->info->opcode == MVM_SSA_PHI) {
+                    location = location->next;
+                }
+                location = location->prev;
+                add_allocation_logging_at_location(tc, g, target, ins, location);
+                break;
+            }
             case MVM_OP_getlex:
             case MVM_OP_getlex_no:
             case MVM_OP_getlexstatic_o:
@@ -249,18 +264,11 @@ void MVM_profile_instrument(MVMThreadContext *tc, MVMStaticFrame *sf) {
 
 /* Ensures we're no longer in instrumented code. */
 void MVM_profile_ensure_uninstrumented(MVMThreadContext *tc, MVMStaticFrame *sf) {
-    if (sf->body.instrumentation && sf->body.bytecode == sf->body.instrumentation->instrumented_bytecode) {
-        /* Switch to uninstrumented code. */
-        sf->body.bytecode      = sf->body.instrumentation->uninstrumented_bytecode;
-        sf->body.handlers      = sf->body.instrumentation->uninstrumented_handlers;
-        sf->body.bytecode_size = sf->body.instrumentation->uninstrumented_bytecode_size;
-
-        /* Throw away specializations, which may also be instrumented. */
-        MVM_spesh_arg_guard_discard(tc, sf);
-
-        /* XXX For now, due to bugs, disable spesh here. */
-        tc->instance->spesh_enabled = 0;
-    }
+    /* XXX due to multithreading trouble, just turning instrumentation off by
+     * switching bytecode back does not work. Profiling instrumentation is
+     * safe to keep around with only a small performance penalty, and CTW
+     * instrumentation is normally not turned off during run time,
+     * so for now we'll just do nothing. */
 }
 
 /* Starts instrumented profiling. */
@@ -315,17 +323,55 @@ typedef struct {
     MVMString *gcs;
     MVMString *time;
     MVMString *full;
+    MVMString *sequence;
+    MVMString *responsible;
     MVMString *cleared_bytes;
     MVMString *retained_bytes;
     MVMString *promoted_bytes;
     MVMString *gen2_roots;
+    MVMString *start_time;
     MVMString *osr;
     MVMString *deopt_one;
     MVMString *deopt_all;
     MVMString *spesh_time;
+    MVMString *thread;
     MVMString *native_lib;
 } ProfDumpStrs;
 
+typedef struct {
+    MVMThreadContext *tc;
+    ProfDumpStrs *pds;
+} ProfTcPdsStruct;
+
+static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds, const MVMProfileCallNode *pcn);
+static MVMObject * dump_call_graph_node_loop(ProfTcPdsStruct *tcpds, const MVMProfileCallNode *pcn) {
+    MVMuint32 i;
+    MVMObject *node_hash;
+
+    node_hash = dump_call_graph_node(tcpds->tc, tcpds->pds, pcn);
+
+    /* Visit successors in the call graph, dumping them and working out the
+     * exclusive time. */
+    if (pcn->num_succ) {
+        MVMObject *callees        = new_array(tcpds->tc);
+        MVMuint64  exclusive_time = pcn->total_time;
+        for (i = 0; i < pcn->num_succ; i++) {
+            MVM_repr_push_o(tcpds->tc, callees,
+                dump_call_graph_node_loop(tcpds, pcn->succ[i]));
+            exclusive_time -= pcn->succ[i]->total_time;
+        }
+        MVM_repr_bind_key_o(tcpds->tc, node_hash, tcpds->pds->exclusive_time,
+            box_i(tcpds->tc, exclusive_time / 1000));
+        MVM_repr_bind_key_o(tcpds->tc, node_hash, tcpds->pds->callees, callees);
+    }
+    else {
+        MVM_repr_bind_key_o(tcpds->tc, node_hash, tcpds->pds->exclusive_time,
+            box_i(tcpds->tc, pcn->total_time / 1000));
+    }
+
+    return node_hash;
+}
+
 /* Dumps a call graph node. */
 static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                                         const MVMProfileCallNode *pcn) {
@@ -407,25 +453,6 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
         MVM_repr_bind_key_o(tc, node_hash, pds->deopt_all,
             box_i(tc, pcn->deopt_all_count));
 
-    /* Visit successors in the call graph, dumping them and working out the
-     * exclusive time. */
-    if (pcn->num_succ) {
-        MVMObject *callees        = new_array(tc);
-        MVMuint64  exclusive_time = pcn->total_time;
-        for (i = 0; i < pcn->num_succ; i++) {
-            MVM_repr_push_o(tc, callees,
-                dump_call_graph_node(tc, pds, pcn->succ[i]));
-            exclusive_time -= pcn->succ[i]->total_time;
-        }
-        MVM_repr_bind_key_o(tc, node_hash, pds->exclusive_time,
-            box_i(tc, exclusive_time / 1000));
-        MVM_repr_bind_key_o(tc, node_hash, pds->callees, callees);
-    }
-    else {
-        MVM_repr_bind_key_o(tc, node_hash, pds->exclusive_time,
-            box_i(tc, pcn->total_time / 1000));
-    }
-
     if (pcn->num_alloc) {
         /* Emit allocations. */
         MVMObject *alloc_list = new_array(tc);
@@ -457,11 +484,21 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
 
 /* Dumps data from a single thread. */
 static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
+                                    MVMThreadContext *othertc,
                                     const MVMProfileThreadData *ptd) {
     MVMObject *thread_hash = new_hash(tc);
     MVMObject *thread_gcs  = new_array(tc);
+    MVMuint64 absolute_start_time;
     MVMuint32  i;
 
+    ProfTcPdsStruct tcpds;
+
+    tcpds.tc = tc;
+    tcpds.pds = pds;
+
+    /* Use the main thread's start time for absolute timings */
+    absolute_start_time = tc->instance->main_thread->prof_data->start_time;
+
     /* Add time. */
     MVM_repr_bind_key_o(tc, thread_hash, pds->total_time,
         box_i(tc, (ptd->end_time - ptd->start_time) / 1000));
@@ -469,7 +506,7 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
     /* Add call graph. */
     if (ptd->call_graph)
         MVM_repr_bind_key_o(tc, thread_hash, pds->call_graph,
-            dump_call_graph_node(tc, pds, ptd->call_graph));
+            dump_call_graph_node_loop(&tcpds, ptd->call_graph));
 
     /* Add GCs. */
     for (i = 0; i < ptd->num_gcs; i++) {
@@ -478,6 +515,10 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
             box_i(tc, ptd->gcs[i].time / 1000));
         MVM_repr_bind_key_o(tc, gc_hash, pds->full,
             box_i(tc, ptd->gcs[i].full));
+        MVM_repr_bind_key_o(tc, gc_hash, pds->sequence,
+            box_i(tc, ptd->gcs[i].gc_seq_num - 1));
+        MVM_repr_bind_key_o(tc, gc_hash, pds->responsible,
+            box_i(tc, ptd->gcs[i].responsible));
         MVM_repr_bind_key_o(tc, gc_hash, pds->cleared_bytes,
             box_i(tc, ptd->gcs[i].cleared_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->retained_bytes,
@@ -486,6 +527,8 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
             box_i(tc, ptd->gcs[i].promoted_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->gen2_roots,
             box_i(tc, ptd->gcs[i].num_gen2roots));
+        MVM_repr_bind_key_o(tc, gc_hash, pds->start_time,
+            box_i(tc, (ptd->gcs[i].abstime - absolute_start_time) / 1000));
         MVM_repr_push_o(tc, thread_gcs, gc_hash);
     }
     MVM_repr_bind_key_o(tc, thread_hash, pds->gcs, thread_gcs);
@@ -494,71 +537,111 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
     MVM_repr_bind_key_o(tc, thread_hash, pds->spesh_time,
         box_i(tc, ptd->spesh_time / 1000));
 
+    /* Add thread id. */
+    MVM_repr_bind_key_o(tc, thread_hash, pds->thread,
+        box_i(tc, othertc->thread_id));
+
     return thread_hash;
 }
 
+void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
+    if (tc->prof_data && tc->prof_data->collected_data) {
+        ProfDumpStrs pds;
+        MVMThread *thread;
+
+        /* We'll allocate the data in gen2, but as we want to keep it, but to be
+         * sure we don't trigger a GC run. */
+        MVM_gc_allocate_gen2_default_set(tc);
+
+        /* Some string constants to re-use. */
+        pds.total_time      = str(tc, "total_time");
+        pds.call_graph      = str(tc, "call_graph");
+        pds.name            = str(tc, "name");
+        pds.id              = str(tc, "id");
+        pds.file            = str(tc, "file");
+        pds.line            = str(tc, "line");
+        pds.entries         = str(tc, "entries");
+        pds.spesh_entries   = str(tc, "spesh_entries");
+        pds.jit_entries     = str(tc, "jit_entries");
+        pds.inlined_entries = str(tc, "inlined_entries");
+        pds.inclusive_time  = str(tc, "inclusive_time");
+        pds.exclusive_time  = str(tc, "exclusive_time");
+        pds.callees         = str(tc, "callees");
+        pds.allocations     = str(tc, "allocations");
+        pds.type            = str(tc, "type");
+        pds.count           = str(tc, "count");
+        pds.spesh           = str(tc, "spesh");
+        pds.jit             = str(tc, "jit");
+        pds.gcs             = str(tc, "gcs");
+        pds.time            = str(tc, "time");
+        pds.full            = str(tc, "full");
+        pds.sequence        = str(tc, "sequence");
+        pds.responsible     = str(tc, "responsible");
+        pds.cleared_bytes   = str(tc, "cleared_bytes");
+        pds.retained_bytes  = str(tc, "retained_bytes");
+        pds.promoted_bytes  = str(tc, "promoted_bytes");
+        pds.gen2_roots      = str(tc, "gen2_roots");
+        pds.start_time      = str(tc, "start_time");
+        pds.osr             = str(tc, "osr");
+        pds.deopt_one       = str(tc, "deopt_one");
+        pds.deopt_all       = str(tc, "deopt_all");
+        pds.spesh_time      = str(tc, "spesh_time");
+        pds.thread          = str(tc, "thread");
+        pds.native_lib      = str(tc, "native library");
+
+        /* Record end time. */
+        tc->prof_data->end_time = uv_hrtime();
+
+        MVM_repr_push_o(tc, tc->prof_data->collected_data, dump_thread_data(tc, &pds, tc, tc->prof_data));
+        while (tc->prof_data->current_call)
+            MVM_profile_log_exit(tc);
+
+        /* Get all thread's data */
+        thread = tc->instance->threads;
+
+        while (thread) {
+            MVMThreadContext *othertc = thread->body.tc;
+            /* Check for othertc to exist because joining threads nulls out
+             * the tc entry in the thread object. */
+            if (othertc && othertc->prof_data && othertc != tc) {
+                /* If we have any call frames still on the profile stack, exit them. */
+                while (othertc->prof_data->current_call)
+                    MVM_profile_log_exit(othertc);
+
+                /* Record end time. */
+                othertc->prof_data->end_time = uv_hrtime();
+
+                MVM_gc_allocate_gen2_default_set(othertc);
+                MVM_repr_push_o(tc, tc->prof_data->collected_data, dump_thread_data(tc, &pds, othertc, othertc->prof_data));
+                MVM_gc_allocate_gen2_default_clear(othertc);
+            }
+            thread = thread->body.next;
+        }
+        MVM_gc_allocate_gen2_default_clear(tc);
+    }
+}
+
 /* Dumps data from all threads into an array of per-thread data. */
 static MVMObject * dump_data(MVMThreadContext *tc) {
-    MVMObject *threads_array;
-    ProfDumpStrs pds;
-
-    /* We'll allocate the data in gen2, but as we want to keep it, but to be
-     * sure we don't trigger a GC run. */
-    MVM_gc_allocate_gen2_default_set(tc);
-
-    /* Some string constants to re-use. */
-    pds.total_time      = str(tc, "total_time");
-    pds.call_graph      = str(tc, "call_graph");
-    pds.name            = str(tc, "name");
-    pds.id              = str(tc, "id");
-    pds.file            = str(tc, "file");
-    pds.line            = str(tc, "line");
-    pds.entries         = str(tc, "entries");
-    pds.spesh_entries   = str(tc, "spesh_entries");
-    pds.jit_entries     = str(tc, "jit_entries");
-    pds.inlined_entries = str(tc, "inlined_entries");
-    pds.inclusive_time  = str(tc, "inclusive_time");
-    pds.exclusive_time  = str(tc, "exclusive_time");
-    pds.callees         = str(tc, "callees");
-    pds.allocations     = str(tc, "allocations");
-    pds.type            = str(tc, "type");
-    pds.count           = str(tc, "count");
-    pds.spesh           = str(tc, "spesh");
-    pds.jit             = str(tc, "jit");
-    pds.gcs             = str(tc, "gcs");
-    pds.time            = str(tc, "time");
-    pds.full            = str(tc, "full");
-    pds.cleared_bytes   = str(tc, "cleared_bytes");
-    pds.retained_bytes  = str(tc, "retained_bytes");
-    pds.promoted_bytes  = str(tc, "promoted_bytes");
-    pds.gen2_roots      = str(tc, "gen2_roots");
-    pds.osr             = str(tc, "osr");
-    pds.deopt_one       = str(tc, "deopt_one");
-    pds.deopt_all       = str(tc, "deopt_all");
-    pds.spesh_time      = str(tc, "spesh_time");
-    pds.native_lib      = str(tc, "native library");
+    MVMObject *collected_data;
 
     /* Build up threads array. */
     /* XXX Only main thread for now. */
-    threads_array = new_array(tc);
-    if (tc->prof_data)
-        MVM_repr_push_o(tc, threads_array, dump_thread_data(tc, &pds, tc->prof_data));
 
-    /* Switch back to default allocation and return result; */
-    MVM_gc_allocate_gen2_default_clear(tc);
-    return threads_array;
+    tc->prof_data->collected_data = new_array(tc);
+
+    /* We rely on the GC orchestration to stop all threads and the
+     * "main" gc thread to dump all thread data for us */
+    MVM_gc_enter_from_allocator(tc);
+
+    collected_data = tc->prof_data->collected_data;
+    tc->prof_data->collected_data = NULL;
+
+    return collected_data;
 }
 
 /* Ends profiling, builds the result data structure, and returns it. */
 MVMObject * MVM_profile_instrumented_end(MVMThreadContext *tc) {
-    if (tc->prof_data) {
-        /* If we have any call frames still on the profile stack, exit them. */
-        while (tc->prof_data->current_call)
-            MVM_profile_log_exit(tc);
-
-        /* Record end time. */
-        tc->prof_data->end_time = uv_hrtime();
-    }
 
     /* Disable profiling. */
     uv_mutex_lock(&(tc->instance->mutex_spesh_sync));
@@ -599,6 +682,8 @@ void MVM_profile_instrumented_mark_data(MVMThreadContext *tc, MVMGCWorklist *wor
                 mark_call_graph_node(tc, node, &nodelist, worklist);
         }
 
+        MVM_gc_worklist_add(tc, worklist, &(tc->prof_data->collected_data));
+
         MVM_free(nodelist.list);
     }
 }
